LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.STD_LOGIC_ARITH.ALL;
USE IEEE.STD_LOGIC_SIGNED.ALL;
PACKAGE coprocessor_parameters IS
  CONSTANT m: INTEGER := 9; -- m rows
  CONSTANT n: INTEGER := 16; -- n columns
  CONSTANT p: INTEGER := 8; -- p-bit data (pixels)
  CONSTANT s: INTEGER := 10; -- s-bit address

  TYPE picture IS ARRAY (0 TO m*n - 1) OF INTEGER;

  CONSTANT unfiltered: picture := (
  1, 1, 2, 3, 4, 5, 6, 7, 8, 7, 6, 5, 4, 3, 2, 1,
  1, 1, 2, 3, 4, 5, 6, 7, 8, 7, 6, 5, 4, 3, 2, 1,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  1, 1, 2, 3, 4, 5, 6, 7, 8, 7, 6, 5, 4, 3, 2, 1,
  1, 1, 2, 3, 4, 5, 6, 7, 8, 7, 6, 5, 4, 3, 2, 1,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  1, 1, 2, 3, 4, 5, 6, 7, 8, 7, 6, 5, 4, 3, 2, 1,
  1, 1, 2, 3, 4, 5, 6, 7, 8, 7, 6, 5, 4, 3, 2, 1,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4
  );

  TYPE mask IS ARRAY (0 TO 8) OF SIGNED(p-1 DOWNTO 0);
  CONSTANT c: mask := (
  x"ff", x"ff", x"ff",
  x"ff",  x"08", x"ff",
  x"ff", x"ff", x"ff"
  );

  TYPE memory IS ARRAY (0 TO 2**s - 1) OF SIGNED(p-1 DOWNTO 0);


END coprocessor_parameters;

LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.STD_LOGIC_ARITH.ALL;
USE IEEE.STD_LOGIC_UNSIGNED.ALL;
USE WORK.coprocessor_parameters.ALL;
ENTITY coprocessor IS
PORT (
    data_in: IN SIGNED(p-1 DOWNTO 0);
    address_in: IN STD_LOGIC_VECTOR(s-1 DOWNTO 0); 
    clk, ADS_in, write_in, Grant: IN STD_LOGIC;
    data_out: OUT SIGNED(p-1 DOWNTO 0);
    address_out: OUT STD_LOGIC_VECTOR(s-1 DOWNTO 0); 
    ADS_out, write_out, Req: OUT STD_LOGIC
);
END coprocessor;

ARCHITECTURE data_flow OF coprocessor IS

  PROCEDURE sync IS BEGIN WAIT UNTIL clk'EVENT AND clk = '1'; END sync;

  SIGNAL a, b: STD_LOGIC_VECTOR(s-1 DOWNTO 0);

  SIGNAL acc: SIGNED(2*p-1 DOWNTO 0);

BEGIN

PROCESS

  VARIABLE pixel_address: INTEGER;

BEGIN

  LOOP

--initial state: bus request signal inactive and
--outputs to the bus in high impedance
     address_out <= (OTHERS => 'Z'); data_out <= (OTHERS => 'Z'); ADS_out <= 'Z'; write_out <= 'Z'; Req <= '0'; sync;

--branch (command decoding)
     IF address_in(9 DOWNTO 7) = "000" AND ADS_in = '1' THEN sync;
       a <= "10"&STD_LOGIC_VECTOR(data_in); sync;
     ELSIF address_in(9 DOWNTO 7) = "001" AND ADS_in = '1' THEN sync;
       b <= "11"&STD_LOGIC_VECTOR(data_in); sync; 
     --request bus control
        Req <= '1'; sync;
        WAIT UNTIL Grant = '1'; sync;
     --filtering operations
        FOR i in 1 TO m-2 LOOP
           FOR j in 1 TO n-2 LOOP
              pixel_address := i*n + j; 
              acc <= x"0000"; sync; 
              data_out <= (OTHERS => 'Z');
              FOR k IN 0 TO 2 LOOP
                 FOR l IN 0 TO 2 LOOP

                 --read data memory (as bus master) and update acc
                 ADS_out <= '1';  write_out <= '0'; 
                 address_out <= a + CONV_STD_LOGIC_VECTOR(pixel_address + (k-1)*n + (l-1), s); sync;
                 ADS_out <= '0'; sync; 
                 acc <= acc + c(k*3 + l)*data_in;

                 END LOOP;
              END LOOP;

              --write data memory (as bus master)
               sync;
               ADS_out <= '1';  write_out <= '1'; 
               address_out <= b + CONV_STD_LOGIC_VECTOR(pixel_address, s); 
               data_out <= acc(p-1 DOWNTO 0); sync;
               ADS_out <= '0'; 

           END LOOP;
        END LOOP;
        sync;
     ELSE sync; 
     END IF;
  END LOOP;
END PROCESS;
END data_flow;



