----------------------------------------------------------------------------
-- Computes: distance = [(x1-x2)^2 + (y1-y2)^2 + (z1-z2)^2]^0.5
-- x1, x2, y1, y2, z1, z2: M bits
-- distance: M+1 bits
----------------------------------------------------------------------------

LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.STD_LOGIC_ARITH.ALL;
USE IEEE.STD_LOGIC_UNSIGNED.ALL;
ENTITY distance_computation IS
  GENERIC(M: NATURAL);
PORT(
  x1, x2, y1, y2, z1, z2: IN STD_LOGIC_VECTOR(M-1 DOWNTO 0);
  clk, reset: IN STD_LOGIC;
  distance: OUT STD_LOGIC_VECTOR(M DOWNTO 0)
);
END distance_computation;

ARCHITECTURE circuit OF distance_computation IS

COMPONENT shift_and_add_multiplier IS
  GENERIC(n: NATURAL:= 64; m: NATURAL:= 64);
PORT(
  x, u: IN STD_LOGIC_VECTOR(n-1 DOWNTO 0);
  y, v: IN STD_LOGIC_VECTOR(m-1 DOWNTO 0);
  clk, reset, start: IN STD_LOGIC;
  z: OUT STD_LOGIC_VECTOR(n+m-1 DOWNTO 0);
  done: OUT STD_LOGIC
);
END COMPONENT;

COMPONENT SquareRoot IS
  GENERIC(n: NATURAL:= 32);
PORT(
  x: IN STD_LOGIC_VECTOR(2*n-1 DOWNTO 0);
  clk, reset, start:IN STD_LOGIC;
  root: OUT STD_LOGIC_VECTOR(n-1 DOWNTO 0);
  remainder: OUT STD_LOGIC_VECTOR(n DOWNTO 0);
  done: OUT STD_LOGIC
);
END COMPONENT;

SIGNAL sel, start1, en, en_pipe, start21, start22: STD_LOGIC;
SIGNAL dif11, dif12, mult1, dif21, dif22, mult2: STD_LOGIC_VECTOR(M-1 DOWNTO 0);
SIGNAL squ11, squ12, squ11d, squ12d,squ2: STD_LOGIC_VECTOR(2*M-1 DOWNTO 0);
SIGNAL sum: STD_LOGIC_VECTOR(2*M+1 DOWNTO 0);
SIGNAL root: STD_LOGIC_VECTOR(M DOWNTO 0);
SIGNAL a_minus_b, c_minus_d: STD_LOGIC_VECTOR(M DOWNTO 0);
SIGNAL b_minus_a, d_minus_c: STD_LOGIC_VECTOR(M-1 DOWNTO 0); 
CONSTANT zero: STD_LOGIC_VECTOR(M-1 DOWNTO 0) := CONV_STD_LOGIC_VECTOR(0, M);

--CONSTANT N: NATURAL := 10;
CONSTANT N: NATURAL := 12;
SUBTYPE state IS NATURAL RANGE 0 TO 2*N+2;
SIGNAL current_state: state;

BEGIN

WITH sel SELECT dif11 <= x1 WHEN '0', y1 WHEN OTHERS;
WITH sel SELECT dif12 <= x2 WHEN '0', y2 WHEN OTHERS;

--differences:
a_minus_b <= ('0'&dif11) - ('0'&dif12);
b_minus_a <= dif12 - dif11;
c_minus_d <= ('0'&dif21) - ('0'&dif22);
d_minus_c <= dif22 - dif21;
WITH a_minus_b(M) SELECT mult1 <= a_minus_b(M-1 DOWNTO 0) WHEN '0', b_minus_a WHEN OTHERS;
WITH c_minus_d(M) SELECT mult2 <= c_minus_d(M-1 DOWNTO 0) WHEN '0', d_minus_c WHEN OTHERS;

first_multiplier: shift_and_add_multiplier
  GENERIC MAP(n => M, m => M)
  PORT MAP(x => mult1, u => zero, y => mult1, v => zero, 
  start => start1, reset => reset, clk => clk, z => squ11);

second_multiplier: shift_and_add_multiplier
  GENERIC MAP(n => M, m => M)
  PORT MAP(x => mult2, u => zero, y => mult2, v => zero, 
  start => start21, reset => reset, clk => clk, z => squ2);

--adder
sum <= (("00"&squ12d) + ("00"&squ11d)) + ("00"&squ2);

square_root: SquareRoot
GENERIC MAP(n => M+1)
PORT MAP(x => sum, clk => clk, start => start22, reset => reset, root => root);


parallel_register: PROCESS(clk)
BEGIN
  IF clk'EVENT AND clk = '1' THEN 
    IF en = '1' THEN squ12 <= squ11; END IF;
  END IF;
END PROCESS;

pipeline_registers: PROCESS(clk)
BEGIN
  IF clk'EVENT AND clk = '1' THEN 
    IF en_pipe = '1' THEN 
      squ12d <= squ12; squ11d <= squ11; 
      dif21 <= z1; dif22 <= z2; distance <= root;
    END IF;
  END IF;
END PROCESS;

--control unit
next_state: PROCESS(clk, reset)
BEGIN
  IF reset = '1' THEN current_state <= 0;
  ELSIF clk'EVENT AND clk = '1' THEN 
    IF current_state < 2*N+2 THEN current_state <= (current_state + 1);
    ELSE current_state <= 0; END IF;
  END IF;
END PROCESS; 

output_function: PROCESS(current_state)
BEGIN
  CASE current_state IS
    WHEN 0 => sel <= '0'; start1 <= '0'; en <= '0'; en_pipe <= '0'; start21 <= '0'; start22 <= '0';

    WHEN 1 => sel <= '0'; start1 <= '1'; en <= '0'; en_pipe <= '0'; start21 <= '1'; start22 <= '0';
    WHEN 2 TO N-1 => sel <= '0'; start1 <= '0'; en <= '0'; en_pipe <= '0'; start21 <= '0'; start22 <= '0';
    WHEN N => sel <= '0'; start1 <= '0'; en <= '1'; en_pipe <= '0'; start21 <= '0'; start22 <= '0';

    WHEN N+1 => sel <= '1'; start1 <= '0'; en <= '0'; en_pipe <= '0'; start21 <= '0'; start22 <= '0';

    WHEN N+2 => sel <= '1'; start1 <= '1'; en <= '0'; en_pipe <= '0'; start21 <= '0'; start22 <= '1';
    WHEN N+3 TO 2*N => sel <= '1'; start1 <= '0'; en <= '0'; en_pipe <= '0'; start21 <= '0'; start22 <= '0';
    WHEN 2*N+1 => sel <= '1'; start1 <= '0'; en <= '0'; en_pipe <= '0'; start21 <= '0'; start22 <= '0';

    WHEN 2*N+2 => sel <= '0'; start1 <= '0'; en <= '0'; en_pipe <= '1'; start21 <= '0'; start22 <= '0';
  END CASE;
END PROCESS;

END circuit;



LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.STD_LOGIC_ARITH.ALL;
USE IEEE.STD_LOGIC_UNSIGNED.ALL;
ENTITY test_distance_computation IS END test_distance_computation;

ARCHITECTURE test OF test_distance_computation IS

COMPONENT distance_computation IS
  GENERIC(M: NATURAL);
PORT(
  x1, x2, y1, y2, z1, z2: IN STD_LOGIC_VECTOR(M-1 DOWNTO 0);
  clk, reset: IN STD_LOGIC;
  distance: OUT STD_LOGIC_VECTOR(M DOWNTO 0)
);
END COMPONENT;

SIGNAL x1, x2, y1, y2, z1, z2: STD_LOGIC_VECTOR(7 DOWNTO 0);
SIGNAL reset: STD_LOGIC;
SIGNAL clk: STD_LOGIC := '0';
SIGNAL distance: STD_LOGIC_VECTOR(8 DOWNTO 0); 

BEGIN

dut: distance_computation
GENERIC MAP(M => 8)
PORT MAP(x1 => x1, x2 => x2, y1 => y1, y2 => y2, z1 => z1, z2 => z2,
clk => clk, reset => reset, distance => distance);

clk <= NOT(clk) AFTER 50 NS;

test_vectors: PROCESS
BEGIN
reset <= '1'; 
x1 <= "00000011"; y1 <= "00000010"; z1 <= "00000001"; 
x2 <= "00000111"; y2 <= "00000110"; z2 <= "00000101"; 
WAIT FOR 100 NS;
reset <= '0';
--WAIT FOR 2700 NS;
--x1 <= "00001011"; y1 <= "00001010"; z1 <= "00001001"; 
--x2 <= "00010111"; y2 <= "00000110"; z2 <= "00000101"; 
--WAIT FOR 2700 NS;
--x1 <= "00000011"; y1 <= "00010010"; z1 <= "00000001"; 
--x2 <= "00000111"; y2 <= "00000110"; z2 <= "00000101"; 
--WAIT FOR 2700 NS;
--x1 <= "00001011"; y1 <= "00001010"; z1 <= "00001001"; 
--x2 <= "00000111"; y2 <= "00100110"; z2 <= "00000101"; 
--WAIT FOR 2700 NS;
--x1 <= "00000000"; y1 <= "00000000"; z1 <= "00000000"; 
--x2 <= "11111111"; y2 <= "11111111"; z2 <= "11111111"; 
--WAIT FOR 2700 NS;
--x1 <= "00000011"; y1 <= "00010010"; z1 <= "00000001"; 
--x2 <= "00000111"; y2 <= "00000110"; z2 <= "00000101"; 
--WAIT FOR 2700 NS;
x1 <= "00001011"; y1 <= "00001010"; z1 <= "00001001"; 
x2 <= "00000111"; y2 <= "00100110"; z2 <= "00000101"; 
WAIT FOR 2700 NS;
x1 <= "00000000"; y1 <= "00000000"; z1 <= "00000000"; 
x2 <= "11111111"; y2 <= "11111111"; z2 <= "11111111"; 
WAIT FOR 2700 NS;
x1 <= "00000011"; y1 <= "00010010"; z1 <= "00000001"; 
x2 <= "00000111"; y2 <= "00000110"; z2 <= "00000101"; 
WAIT FOR 2700 NS;
x1 <= "00001011"; y1 <= "00001010"; z1 <= "00001001"; 
x2 <= "00000111"; y2 <= "00100110"; z2 <= "00000101"; 
WAIT FOR 2700 NS;
x1 <= "00000000"; y1 <= "00000000"; z1 <= "00000000"; 
x2 <= "11111111"; y2 <= "11111111"; z2 <= "11111111"; 
WAIT;
END PROCESS;

END test;