----------------------------------------------------------------------------
--parameterized FIR filter
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_signed.all;
package fir_parameters is
  constant k: integer := 10;
  constant m: integer := 24;
  type coefficients is array (0 to k) of signed(m-1 downto 0);

  constant a: coefficients := (
    x"fd1e75",
    x"000000",
    x"04cd3a",  
    x"0a2f98",  
    x"0e67b2",  
    x"100000",  
    x"0e67b2",  
    x"0a2f98",  
    x"04cd3a",  
    x"000000",  
    x"fd1e75"
  );

end fir_parameters;

library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_signed.all;
use work.fir_parameters.all;

entity fir_filter is
port (
  x: in signed(m-1 downto 0);
  clk, reset: in std_logic;
  y: out signed(m-1 downto 0)
);
end fir_filter;

architecture circuit of fir_filter is

  type cell_outputs is array (0 to k) of signed(m-1 downto 0);
  type unrounded is array (0 to k) of signed(2*m-1 downto 0);  
  signal acc, register_out, mac_out: cell_outputs;
  signal product: unrounded;

begin

   acc(k) <= conv_signed(0, m);
   product(k) <= a(k)*x;
   mac_out(k) <= acc(k) + product(k)(2*m-3 downto m-2);

   registers_k: process(clk, reset)
   begin
     if reset = '1' then register_out(k) <= conv_signed(0, m);
     elsif clk'event and clk = '1' then register_out(k) <= mac_out(k);
     end if;
   end process;

   iterative_cell: for j in 0 to k-1 generate 
      acc(j) <= register_out(j+1);
      product(j) <= a(j)*x;   
      mac_out(j) <= acc(j) + product(j)(2*m-3 downto m-2);
      registers: process(clk, reset)
      begin
        if reset = '1' then register_out(j) <= conv_signed(0, m);
        elsif clk'event and clk = '1' then register_out(j) <= mac_out(j);
        end if;
      end process;
    end generate;
    
    y <= register_out(0);

end circuit;


library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_signed.all;
use work.fir_parameters.all;

entity test_fir is end test_fir;

architecture test of test_fir is

component fir_filter is
port (
  x: in signed(m-1 downto 0);
  clk, reset: in std_logic;
  y: out signed(m-1 downto 0)
);
end component;

signal x, y: signed(m-1 downto 0);
signal reset: std_logic;
signal clk: std_logic := '0';
signal synch: std_logic := '0';
signal analog_in, analog_out: real;

begin

dut: fir_filter
port map( x => x, y => y, clk => clk, reset => reset);

clk <= not(clk) after 50 ns;

reset <= '1', '0' after 100 ns;

--step:
--x <= x"c00000", x"400000" after 200 ns;

--impulse:
x <= x"000000", x"400000" after 100 ns, x"000000" after 200 ns;

--periodic signal:
--synch <= not(synch) after 100 ns;
--with synch select x <= x"c00000" when '0', x"400000" when others;

dac: process(x, y)
variable a, b: signed(m-1 downto 0);
variable c, d: real;
begin
a := y; b := x;
if a(m-1) = '1' then a(m-1) := '0';
c := real(conv_integer(a));
c := c / real((2**22));
c := c - 2.0;
else 
c := real(conv_integer(a));
c := c / real((2**22));
end if;
if b(m-1) = '1' then b(m-1) := '0';
d := real(conv_integer(b));
d := d / real((2**22));
d := d - 2.0;
else 
d := real(conv_integer(b));
d := d / real((2**22));
end if;

analog_in <= d; analog_out <= c;


end process;

end test;


