LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.STD_LOGIC_ARITH.ALL;
USE IEEE.STD_LOGIC_UNSIGNED.ALL;
ENTITY logarithm_cell_DS IS 
   GENERIC(m: natural);
PORT(
   z: in std_logic_vector(m+3 downto 0);
   next_z: out std_logic_vector(m+3 downto 0);
   t: out std_logic_vector(0 to 1)
);
END logarithm_cell_DS;

ARCHITECTURE circuit OF logarithm_cell_DS IS
   SIGNAL trunc, trunc_div_8, trunc_div_4, trunc_div_2: std_logic_vector(m+3 downto 0);
   SIGNAL mult_in: std_logic_vector(m downto 0);
   SIGNAL mult_1: std_logic_vector(2*m+1 downto 0);
   SIGNAL mult_2: std_logic_vector(4*m+3 downto 0);
BEGIN

mult_in <= z(m downto 0);
mult_1 <= mult_in * mult_in;
mult_2 <= mult_1 * mult_1;

trunc <= mult_2(4*m+3 downto 3*m);
trunc_div_2 <= '0'&trunc(m+3 downto 1);
trunc_div_4 <= "00"&trunc(m+3 downto 2);
trunc_div_8 <= "000"&trunc(m+3 downto 3);

selector: PROCESS(trunc, trunc_div_2, trunc_div_4, trunc_div_8)
BEGIN
   IF trunc(m+3) = '1' THEN next_z <= trunc_div_8; t <= "11";
   ELSIF trunc(m+2) = '1' THEN next_z <= trunc_div_4; t <= "10";
   ELSIF trunc(m+1) = '1' THEN next_z <= trunc_div_2; t <= "01";
   ELSE next_z <= trunc; t <= "00";
   END IF;
END PROCESS;

END circuit;


LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.STD_LOGIC_ARITH.ALL;
USE IEEE.STD_LOGIC_UNSIGNED.ALL;
ENTITY logarithm_pipe_DS IS 
   GENERIC(n, m, p: natural);
PORT(
   x: in std_logic_vector(0 to n);
   clk: in std_logic;
   y: out std_logic_vector(1 to p)
);
END logarithm_pipe_DS;

ARCHITECTURE circuit OF logarithm_pipe_DS IS
   
   COMPONENT logarithm_cell_DS IS 
   GENERIC(m: natural);
   PORT(
   z: in std_logic_vector(m+3 downto 0);
   next_z: out std_logic_vector(m+3 downto 0);
   t: out std_logic_vector(0 to 1)
   );
   END COMPONENT;

   TYPE vector_data IS ARRAY (1 TO p/2) OF std_logic_vector(m+3 downto 0);
   SIGNAL vector_z, pipe: vector_data;

   TYPE quaternary_vector IS ARRAY (1 TO p/2) OF STD_LOGIC_VECTOR(0 TO 1);    
   TYPE vector_out IS ARRAY (1 TO p/2) OF quaternary_vector;
   SIGNAL yy: vector_out;
   SIGNAL tt: quaternary_vector;

--   TYPE vector_out IS ARRAY (1 TO p/2) OF std_logic_vector(1 to p);
--   SIGNAL yy: vector_out;
--   SIGNAL tt: std_logic_vector(1 to p);

   CONSTANT zeroes: std_logic_vector(m-n-1 downto 0) := (others => '0'); 

BEGIN

vector_z(1) <= "000"&x&zeroes;
iteration: FOR i IN 1 TO p/2-1 GENERATE
   iterative_cell: logarithm_cell_DS
   GENERIC MAP (m => m)
--   PORT MAP(z => vector_z(i), next_z => pipe(i), t => tt(2*i-1 to 2*i));
   PORT MAP(z => vector_z(i), next_z => pipe(i), t => tt(i));

   pipe_registers: PROCESS(clk)
   BEGIN
      IF clk'EVENT AND clk = '1' THEN vector_z(i+1) <= pipe(i);
--         yy(i)<= tt(2*i-1 to 2*i) & yy(i)(1 to p-2);
         yy(i)<= tt(i) & yy(i)(1 to p/2-1);
      END IF;
   END PROCESS;
   END GENERATE;

last_cell: logarithm_cell_DS
   GENERIC MAP (m => m)
--   PORT MAP(z => vector_z(p/2), t => tt(p-1 to p));
   PORT MAP(z => vector_z(p/2), t => tt(p/2));
   last_register: PROCESS(clk)
   BEGIN
      IF clk'EVENT AND clk = '1' THEN 
--         yy(p/2)<= tt(p-1 to p) & yy(p/2)(1 to p-2);
         yy(p/2)<= tt(p/2) & yy(p/2)(1 to p/2-1);
      END IF;
   END PROCESS;

output_generation: PROCESS(yy)
   variable i: natural;
BEGIN
--   FOR i IN 1 TO p LOOP
--      if i mod 2 = 1 then 
--      y(i) <= yy((i+1)/2)(p-i);
--      else y(i) <= yy(i/2)(p-i+2);
--      end if;
--   END LOOP;
   FOR i IN 1 TO p/2 LOOP
      y(2*i-1 to 2*i) <= yy(i)(p/2-i+1);
   END LOOP;
END PROCESS;

END circuit;


LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.STD_LOGIC_ARITH.ALL;
USE IEEE.STD_LOGIC_UNSIGNED.ALL;
ENTITY test_logarithm_pipeline_DS IS END test_logarithm_pipeline_DS;

ARCHITECTURE test OF test_logarithm_pipeline_DS IS

COMPONENT logarithm_pipe_DS IS 
   GENERIC(n, m, p: natural);
PORT(
   x: in std_logic_vector(0 to n);
   clk: in std_logic;
   y: out std_logic_vector(1 to p)
);
END COMPONENT;

   SIGNAL x: std_logic_vector(0 to 8);
   SIGNAL y: std_logic_vector(1 to 16);
   SIGNAL clk: STD_LOGIC := '0';
BEGIN

  dut: logarithm_pipe_DS
     generic map(n => 8, m => 16, p => 16)
     port map(x => x, y => y, clk => clk);

   clk <= NOT(clk) AFTER 50 NS;

test_vectors: PROCESS
BEGIN
   x <= "110110001";
   WAIT FOR 50 NS;
   x <= "100110001";
   WAIT FOR 100 NS;
   x <= "111110001";
   WAIT FOR 100 NS;
   x <= "100000101";
   WAIT FOR 100 NS;
   x <= "110110001";
   WAIT FOR 100 NS;
   x <= "100110001";
   WAIT FOR 100 NS;
   x <= "111110001";
   WAIT FOR 100 NS;
   x <= "100000101";
   WAIT FOR 100 NS;
   x <= "110110001";
   WAIT FOR 100 NS;
   x <= "100110001";
   WAIT FOR 100 NS;
   x <= "111110001";
   WAIT FOR 100 NS;
   x <= "100000101";
   WAIT;

END PROCESS;

END test;
