LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.STD_LOGIC_ARITH.ALL;
USE IEEE.STD_LOGIC_UNSIGNED.ALL;
ENTITY logarithm_cell IS 
   GENERIC(m: natural);
PORT(
   z: in std_logic_vector(m+1 downto 0);
   next_z: out std_logic_vector(m+1 downto 0);
   t: out std_logic
);
END logarithm_cell;

ARCHITECTURE circuit OF logarithm_cell IS
   SIGNAL trunc_square, trunc_div: std_logic_vector(m+1 downto 0);
   SIGNAL mult_in: std_logic_vector(m downto 0);
   SIGNAL mult_out: std_logic_vector(2*m+1 downto 0);
BEGIN

WITH trunc_square(m+1) SELECT next_z <= trunc_square WHEN '0', 
                              trunc_div WHEN others;         

mult_in <= z(m downto 0);
mult_out <= mult_in * mult_in;
trunc_square <= mult_out(2*m+1 downto m);
t <= trunc_square(m+1);
trunc_div <= '0'&trunc_square(m+1 downto 1);

END circuit;



LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.STD_LOGIC_ARITH.ALL;
USE IEEE.STD_LOGIC_UNSIGNED.ALL;
ENTITY logarithm_pipeline IS 
   GENERIC(n, m, p: natural);
PORT(
   x: in std_logic_vector(0 to n);
   clk: in std_logic;
   y: out std_logic_vector(1 to p)
);
END logarithm_pipeline;

ARCHITECTURE circuit OF logarithm_pipeline IS
    
   COMPONENT logarithm_cell IS 
   GENERIC(m: natural);
   PORT(
   z: in std_logic_vector(m+1 downto 0);
   next_z: out std_logic_vector(m+1 downto 0);
   t: out std_logic
   );
   END COMPONENT;

   TYPE vector_data IS ARRAY (1 TO p) OF std_logic_vector(m+1 downto 0);
   SIGNAL vector_z, pipe: vector_data;
   TYPE vector_out IS ARRAY (1 TO p) OF std_logic_vector(1 to p);
   SIGNAL yy: vector_out;
   SIGNAL tt: std_logic_vector(1 to p);
   CONSTANT zeroes: std_logic_vector(m-n-1 downto 0) := (others => '0'); 
   
BEGIN

vector_z(1) <= '0'&x&zeroes;
iteration: FOR i IN 1 TO p-1 GENERATE
   iterative_cell: logarithm_cell
   GENERIC MAP (m => m)
   PORT MAP(z => vector_z(i), next_z => pipe(i), t => tt(i));
   pipe_registers: PROCESS(clk)
      BEGIN
         IF clk'EVENT AND clk = '1' THEN vector_z(i+1) <= pipe(i);
            yy(i)<= tt(i) & yy(i)(1 to p-1);
         END IF;
      END PROCESS;
   END GENERATE;

last_cell: logarithm_cell
   GENERIC MAP (m => m)
   PORT MAP(z => vector_z(p), t => tt(p));
   last_register: PROCESS(clk)
      BEGIN
         IF clk'EVENT AND clk = '1' THEN 
            yy(p)<= tt(p) & yy(p)(1 to p-1);
         END IF;
      END PROCESS;

output_generation: PROCESS(yy)
BEGIN
   FOR i IN 1 TO p LOOP
      y(i) <= yy(i)(p-i+1);
   END LOOP;
END PROCESS;

END circuit;


LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.STD_LOGIC_ARITH.ALL;
USE IEEE.STD_LOGIC_UNSIGNED.ALL;
ENTITY test_logarithm_pipeline IS END test_logarithm_pipeline;

ARCHITECTURE test OF test_logarithm_pipeline IS

COMPONENT logarithm_pipeline IS 
   GENERIC(n, m, p: natural);
PORT(
   x: in std_logic_vector(0 to n);
   clk: in std_logic;
   y: out std_logic_vector(1 to p)
);
END COMPONENT;

   SIGNAL x: std_logic_vector(0 to 8);
   SIGNAL y: std_logic_vector(1 to 16);
   SIGNAL clk: STD_LOGIC := '0';
BEGIN

  dut: logarithm_pipeline
     generic map(n => 8, m => 16, p => 16)
     port map(x => x, y => y, clk => clk);

   clk <= NOT(clk) AFTER 50 NS;

test_vectors: PROCESS
BEGIN
   x <= "110110001";
   WAIT FOR 50 NS;
   x <= "100110001";
   WAIT FOR 100 NS;
   x <= "111110001";
   WAIT FOR 100 NS;
   x <= "100000101";
   WAIT FOR 100 NS;
   x <= "110110001";
   WAIT FOR 100 NS;
   x <= "100110001";
   WAIT FOR 100 NS;
   x <= "111110001";
   WAIT FOR 100 NS;
   x <= "100000101";
   WAIT FOR 100 NS;
   x <= "110110001";
   WAIT FOR 100 NS;
   x <= "100110001";
   WAIT FOR 100 NS;
   x <= "111110001";
   WAIT FOR 100 NS;
   x <= "100000101";
   WAIT;

END PROCESS;

END test;
