library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package pid_parameters is
  constant m: integer := 16; --number of bits of memory word
  constant n: integer := 8;  --number of memory words
  constant p: integer := 3;  --number of address bits
  constant k1: std_logic_vector(m-1 downto 0) := x"006b";
  constant k2: std_logic_vector(m-1 downto 0) := x"ff98";
  constant k3: std_logic_vector(m-1 downto 0) := x"0002";
  constant minus_one: std_logic_vector(m-1 downto 0) := x"ffff"; 
  constant zero: std_logic_vector(m-1 downto 0) := x"0000"; 
  constant sample_period: integer := 16; --number of clock cycles 
end pid_parameters;

library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.pid_parameters.all;
entity three_port_register_file is
port (
  data_in: in std_logic_vector(m-1 downto 0);
  clk, write_reg, reset: in std_logic;
  address_A, address_B, address_W: in std_logic_vector(p-1 downto 0);
  data_out_A, data_out_B: out std_logic_vector(m-1 downto 0)  
);
end three_port_register_file;

architecture structure of three_port_register_file is

type memory is array (0 to n-1) of std_logic_vector(m-1 downto 0);
signal X: memory;
signal EN: std_logic_vector(0 to n-1);

begin

  decoder: process(address_W, write_reg)
  begin
    for i in 0 to n-1 loop
      if i = conv_integer(address_W) then EN(i) <= write_reg;
      else EN(i) <= '0';
      end if;
    end loop;
  end process;
  
   bank_registers: process(clk, reset)
   begin
     if reset = '1' THEN
       for i in 0 to n-1 loop X(i) <= (OTHERS => '0'); end loop;
     elsif clk'event and clk = '1' then
       for i in 0 to n-1 loop
         if EN(i) = '1' then X(i) <= data_in; end if;
       end loop;
     end if;
    end process;

  multiplexer_A: process(address_A, X)
  begin
     for i in 0 to n-1 loop
        if i = conv_integer(address_A) then data_out_A <= X(i); end if;
     end loop;
   end process;  

  multiplexer_B: process(address_B, X)
  begin
     for i in 0 to n-1 loop
        if i = conv_integer(address_B) then data_out_B <= X(i); end if;
     end loop;
   end process;  


end structure;	

library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.pid_parameters.all;
entity computation_resource is
port (
  x: in std_logic_vector(m-1 downto 0);
  y: in std_logic_vector(m-1 downto 0);
  oper: in std_logic_vector(2 downto 0);
  z: out std_logic_vector(m-1 downto 0)
);
end computation_resource;

architecture circuit of computation_resource is

component modified_parallel_multiplier IS
  GENERIC(n: NATURAL:= 8; m: NATURAL:= 16);
PORT(
  x, u: IN STD_LOGIC_VECTOR(n DOWNTO 0);
  y: IN STD_LOGIC_VECTOR(m DOWNTO 0);
  v: IN STD_LOGIC_VECTOR(m-1 DOWNTO 0);
  z: OUT STD_LOGIC_VECTOR(n+m+1 DOWNTO 0)
);
END component;

signal mux_out: std_logic_vector(m-1 downto 0);
signal long_z: std_logic_vector(2*m-1 downto 0);

begin
   with oper select mux_out <=  minus_one when "001", 
                          k1 when "010",
                          k2 when "100",
                          k3 when "110",
                          zero when others;

   main_component: modified_parallel_multiplier
      generic map(n => m-1, m => m-1)
      port map(x => mux_out, y => y, u => x, v => zero(m-2 downto 0), z => long_z);

  z <= long_z(m-1 downto 0);
end circuit;

library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.pid_parameters.all;
entity pid_controller2 is
port (
  reference, s: in std_logic_vector(m-1 downto 0);
  clk, reset: in std_logic;
  u: out std_logic_vector(m-1 downto 0)
);
end pid_controller2;

architecture circuit of pid_controller2 is

component computation_resource is
port (
  x: in std_logic_vector(m-1 downto 0);
  y: in std_logic_vector(m-1 downto 0);
  oper: in std_logic_vector(2 downto 0);
  z: out std_logic_vector(m-1 downto 0)
);
end component;

component three_port_register_file is
port (
  data_in: in std_logic_vector(m-1 downto 0);
  clk, write_reg, reset: in std_logic;
  address_A, address_B, address_W: in std_logic_vector(p-1 downto 0);
  data_out_A, data_out_B: out std_logic_vector(m-1 downto 0)  
);
end component;

signal x, z: std_logic_vector(m-1 downto 0);
signal y: std_logic_vector(m-1 downto 0);
signal data_out_A, data_out_B: std_logic_vector(m-1 downto 0);
signal sel, out_en, write_reg, time_out: std_logic;
signal oper: std_logic_vector(2 downto 0);
signal address_A, address_B, address_W: std_logic_vector(p-1 downto 0);
signal count: natural;

SUBTYPE state IS NATURAL RANGE 0 TO 6;
SIGNAL current_state: state;

begin

---DATA PATH
first_component: computation_resource
port map (x => x, y => y, oper => oper, z => z);

second_component: three_port_register_file
port map (data_in => z, address_A => address_A,  address_B => address_B, address_W => address_W, 
write_reg => write_reg, clk => clk, reset => reset, data_out_A => data_out_A,  data_out_B => data_out_B);

WITH sel SELECT x <= reference WHEN '0', data_out_A WHEN OTHERS;
WITH sel SELECT y <= s WHEN '0', data_out_B WHEN OTHERS;

output_register: process(clk)
BEGIN
   IF clk'event AND clk = '1' THEN
     IF out_en = '1' THEN u <= z; END IF;
   END IF;
END PROCESS;

---TIMER
counter: PROCESS(clk, reset)
BEGIN
   IF reset = '1' THEN count <= sample_period - 1;
   ELSIF clk'event AND clk = '1' THEN count <= (count - 1) MOD sample_period;
   END IF;
END PROCESS;
WITH count SELECT time_out <= '1' WHEN 0, '0' WHEN OTHERS;

---CONTROL UNIT
next_state_function: PROCESS(clk, reset)
BEGIN
   IF reset = '1' THEN current_state <= 0;
   ELSIF clk'event AND clk = '1' THEN
      CASE current_state IS
         WHEN 0 TO 5 => current_state <= current_state + 1;
         WHEN OTHERS => IF time_out = '1' THEN current_state <= 0; END IF;
      END CASE;
   END IF;
END PROCESS;

output_function: PROCESS(current_state)
BEGIN
      CASE current_state IS
         WHEN 0 => write_reg <= '1'; sel <= '0'; address_W <= "000"; 
         address_A <= "000"; address_B <= "000"; oper <= "001"; out_en <= '0';
         WHEN 1 => write_reg <= '1'; sel <= '1'; address_W <= "100"; 
         address_A <= "011"; address_B <= "000"; oper <= "010"; out_en <= '0';
         WHEN 2 => write_reg <= '1'; sel <= '1'; address_W <= "100"; 
         address_A <= "100"; address_B <= "001"; oper <= "100"; out_en <= '0';
         WHEN 3 => write_reg <= '1'; sel <= '1'; address_W <= "011"; 
         address_A <= "100"; address_B <= "010"; oper <= "110"; out_en <= '1';
         WHEN 4 => write_reg <= '1'; sel <= '1'; address_W <= "010"; 
         address_A <= "001"; address_B <= "000"; oper <= "000"; out_en <= '0';
         WHEN 5 => write_reg <= '1'; sel <= '1'; address_W <= "001"; 
         address_A <= "000"; address_B <= "000"; oper <= "000"; out_en <= '0';
         WHEN OTHERS => write_reg <= '0'; sel <= '0'; address_W <= "000"; 
         address_A <= "000"; address_B <= "000"; oper <= "000"; out_en <= '0';
      END CASE;
END PROCESS;

END circuit;


library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.pid_parameters.all;
entity test_pid_controller2 IS END test_pid_controller2;

architecture test of test_pid_controller2 is

component pid_controller2 is
port (
  reference, s: in std_logic_vector(m-1 downto 0);
  clk, reset: in std_logic;
  u: out std_logic_vector(m-1 downto 0)
);
end component;

signal reference, s: std_logic_vector(m-1 downto 0);
signal u: std_logic_vector(m-1 downto 0); 
signal reset : STD_LOGIC;
signal clk: STD_LOGIC := '0';

BEGIN

dut: pid_controller2
port map(reference => reference, s => s, u => u, reset => reset, clk => clk);

clk <= NOT(clk) AFTER 50 NS;

test_vectors: PROCESS
BEGIN
reset <= '1';
reference <= x"0007";
s <= x"000f";
WAIT FOR 100 NS;
reset <= '0';
WAIT FOR 1000 NS;
s <= x"0005";
WAIT FOR 1600 NS;
s <= x"0008";
WAIT FOR 1600 NS;
s <= x"0007";
WAIT;
END PROCESS;
END test;