library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package main_parameters is
  constant m: natural := 8; -- m bits per word
  constant n: natural := 16; -- n words
  constant p: natural := 4; -- p address bits
  constant zero: std_logic_vector(m-1 downto 0) := conv_std_logic_vector(0, m);
end main_parameters;

library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.main_parameters.all;
entity two_port_register_file is
port (
  data_in_A, data_in_B: in std_logic_vector(m-1 downto 0);
  clk, write_A, write_B, read_A, read_B: in std_logic;
  address_A, address_B: in std_logic_vector(p-1 downto 0);
  data_out_A, data_out_B: out std_logic_vector(m-1 downto 0)  
);
end two_port_register_file;

architecture structure of two_port_register_file is

type memory is array (0 to n-1) of std_logic_vector(m-1 downto 0);
signal X, reg_in: memory;
signal EN_A, EN_B: std_logic_vector(0 to n-1);
signal reg_out_A, reg_out_B: std_logic_vector(m-1 downto 0); 

begin

  decoder_A: process(address_A, write_A)
  begin
    for i in 0 to n-1 loop
      if i = conv_integer(address_A) then EN_A(i) <= write_A;
      else EN_A(i) <= '0';
      end if;
    end loop;
  end process;
  
  decoder_B: process(address_B, write_B)
  begin
    for i in 0 to n-1 loop
      if i = conv_integer(address_B) then EN_B(i) <= write_B;
      else EN_B(i) <= '0';
      end if;
    end loop;
  end process;

---------------------------

  register_inputs: for i in 0 to n-1 generate
     with EN_B(i) select reg_in(i) <= data_in_A when '0', data_in_B when others;
  end generate;

-------------------

   bank_registers: process(clk)
   begin
     if clk'event and clk = '1' then
       for i in 0 to n-1 loop
         if (EN_A(i) = '1') OR (EN_B(i) = '1') then X(i) <= reg_in(i); end if;
       end loop;
     end if;
    end process;

  multiplexer_A: process(address_A, X)
  begin
     for i in 0 to n-1 loop
        if i = conv_integer(address_A) then  reg_out_A <= X(i); end if;
     end loop;
   end process;  

  multiplexer_B: process(address_B, X)
  begin
     for i in 0 to n-1 loop
        if i = conv_integer(address_B) then  reg_out_B <= X(i); end if;
     end loop;
   end process;  

  output_register_A: process(clk)
  begin
     if clk'event and clk = '1' then
        if read_A = '1' then data_out_A <= reg_out_A; end if;
     end if;
  end process;

  output_register_B: process(clk)
  begin
     if clk'event and clk = '1' then
        if read_B = '1' then data_out_B <= reg_out_B; end if;
     end if;
  end process;

	
end structure;	


library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.main_parameters.all;
entity test_two_port_register_file is end test_two_port_register_file;

architecture test of test_two_port_register_file is

component two_port_register_file is
port (
  data_in_A, data_in_B: in std_logic_vector(m-1 downto 0);
  clk, write_A, write_B, read_A, read_B: in std_logic;
  address_A, address_B: in std_logic_vector(p-1 downto 0);
  data_out_A, data_out_B: out std_logic_vector(m-1 downto 0)  
);
end component;

signal data_in_A, data_in_B: std_logic_vector(m-1 downto 0);
signal clk: std_logic := '0';
signal write_A, write_B, read_A, read_B: std_logic;
signal address_A, address_B: std_logic_vector(p-1 downto 0);
signal data_out_A, data_out_B: std_logic_vector(m-1 downto 0);

begin

dut: two_port_register_file
port map(data_in_A => data_in_A, data_in_B => data_in_B, clk => clk, 
write_A => write_A, write_B => write_B, read_A => read_A, read_B => read_B, 
address_A => address_A, address_B => address_B,  
data_out_A => data_out_A, data_out_B => data_out_B);
  
clk <= not(clk) after 10 ns;

process
begin
write_A <= '1'; write_B <= '1';
read_A <= '1'; read_B <= '1';
data_in_A <= "00000000"; data_in_B <= "00000011"; 
address_A <= "0000"; address_B <= "1000";

for j in 1 to n/2-1 loop
  wait for 100 ns;
  address_A <= address_A + 1;
  data_in_A <= data_in_A + 1;
  address_B <= address_B + 1;
  data_in_B <= data_in_B + 1;
end loop;

wait for 100 ns;
write_A <= '0'; write_B <= '0';
wait for 100 ns;
read_A <= '0'; read_B <= '0';
wait for 100 ns;

address_A <= "0000"; address_B <= "0100"; 
read_A <= '1'; read_B <= '1';

for j in 0 to n-1 loop
  wait for 100 ns;
  address_A <= address_A + 1; address_B <= address_B + 1; 
end loop;

wait;

end process;

end test;

